from dataclasses import dataclass
from typing import Callable, List, Union
import numpy as np

# Tensorflow is extraordinarily noisy. Catch warnings during import.
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import tensorflow as tf

from .op import Op
from .metric import Euclidean, ed
from .param import get_parameter_values, ppp, upp, bpp

@dataclass
class GP(Op):
    def __init__(self, fa, autoinputs):
        if 'offset' not in autoinputs: autoinputs['offset'] = 'offset'
        if 'locs1' not in autoinputs: autoinputs['locs1'] = 'locs1'
        if 'locs2' not in autoinputs: autoinputs['locs2'] = 'locs2'
        super().__init__(fa, autoinputs)

    def __add__(self, other):
        return Stack([self]) + other

    def __mul__(self, other):
        return Product([self]) * other

    def call(self, p, e):
        """
        Returns tuple `(mean, covariance)` for locations.
        Return values may be unbroadcasted.
        """
        pass

    def __call__(self, p, e):
        """
        Returns tuple `(mean, covariance)` for locations.
        Return values have correct shapes.
        """
        M, C = self.call(p, e)
        if M is None: M = 0.
        if C is None: C = 0.
        n1 = tf.shape(e['locs1'])[0]
        n2 = tf.shape(e['locs2'])[0]
        M = tf.broadcast_to(M, [n1])
        C = tf.broadcast_to(C, [n1, n2])
        return M, C

    def report(self, p):
        string = ', '.join('%s %4.2f' % (v.name, p[v.name]) for v in self.vars())
        return '[' + string + ']'

    def reg(self, p):
        pass

def get_trend_coefs(beta):
    if isinstance(beta, (list, tuple)):
        return [p for s in beta for p in upp(s)]
    elif isinstance(beta, str):
        return upp(beta)
    else:
        return []

class Trend(GP):
    def __init__(self, featurizer, beta='beta'):
        fa = dict(beta=beta)
        self.featurizer = featurizer
        super().__init__(fa, dict(locs1='locs1'))

    def vars(self):
        return get_trend_coefs(self.fa['beta'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        x = tf.cast(self.featurizer(e['locs1']), tf.float32)
        if isinstance(v['beta'], (tuple, list)):
            v['beta'] = tf.stack(v['beta'])
        return tf.einsum('ab,b->a', x, v['beta']), None # [locs1]

    def reg(self, p):
        return 0.

class TrendPrior(GP):
    def __init__(self, featurizer, alpha='alpha'):
        fa = dict(alpha=alpha)
        self.featurizer = featurizer
        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['alpha'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        F1 = tf.cast(self.featurizer(e['locs1']), tf.float32)
        F2 = tf.cast(self.featurizer(e['locs2']), tf.float32)
        return None, v['alpha'] * tf.einsum('ba,ca->bc', F1, F2)

    def reg(self, p):
        return 0.

def scale_to_metric(scale, metric):
    assert scale is None or metric is None
    if metric is None:
        if scale is None:
            metric = 'euclidean'
        else:
            metric = Euclidean(scale)
    return metric

class SquaredExponential(GP):
    def __init__(self, sill='sill', range='range', scale=None, metric=None):
        fa = dict(sill=sill, range=range)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp(self.fa['sill']) + ppp(self.fa['range'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        return None, v['sill'] * tf.exp(-0.5 * e['d2'] / tf.square(v['range']))

    def reg(self, p):
        v = get_parameter_values(self.fa, p)
        return v['range']

class GammaExponential(GP):
    def __init__(self, range='range', sill='sill', gamma='gamma', scale=None, metric=None):
        fa = dict(sill=sill, range=range, gamma=gamma, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp(self.fa['sill']) + ppp(self.fa['range']) + bpp(self.fa['gamma'], 0., 2.)

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        return None, v['sill'] * gamma_exp(e['d2'] / tf.square(v['range']), v['gamma'])

    def reg(self, p):
        v = get_parameter_values(self.fa, p)
        return v['range']

class Wiener(GP):
    def __init__(self, axis, start):

        self.axis = axis
        self.start = start

        # Include the element of scale corresponding to the axis of
        # integration as an explicit formal argument.
        fa = dict()

        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return []

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        x1 = e['locs1'][..., self.axis]
        x2 = e['locs2'][..., self.axis]
        k = tf.minimum(ed(x1, 1), ed(x2, 0)) - self.start
        return None, k

    def reg(self, p):
        return 0.

class IntSquaredExponential(GP):
    def __init__(self, axis, start, range='range'):

        self.axis = axis
        self.start = start

        # Include the element of scale corresponding to the axis of
        # integration as an explicit formal argument.
        fa = dict(range=range)

        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['range'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        x1 = tf.pad(e['locs1'][..., self.axis] - self.start, [[1, 0]])
        x2 = tf.pad(e['locs2'][..., self.axis] - self.start, [[1, 0]])

        r = v['range']
        sdiff = (ed(x1, 1) - ed(x2, 0)) / (r * np.sqrt(2.))
        k = -tf.square(r) * (np.sqrt(np.pi) * sdiff * tf.math.erf(sdiff) + tf.exp(-tf.square(sdiff)))
        k -= k[0:1, :]
        k -= k[:, 0:1]
        k = k[1:, 1:]

        return None, k

    def reg(self, p):
        return 0.

class IntExponential(GP):
    def __init__(self, axis, start, range='range'):

        self.axis = axis
        self.start = start

        # Include the element of scale corresponding to the axis of
        # integration as an explicit formal argument.
        fa = dict(range=range)

        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['range'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        x1 = tf.pad(e['locs1'][..., self.axis] - self.start, [[1, 0]])
        x2 = tf.pad(e['locs2'][..., self.axis] - self.start, [[1, 0]])

        r = v['range']
        sdiff = tf.abs(ed(x1, 1) - ed(x2, 0)) / r
        k = -tf.square(r) * (sdiff + tf.exp(-sdiff))
        k -= k[0:1, :]
        k -= k[:, 0:1]
        k = k[1:, 1:]

        return None, k

    def reg(self, p):
        return 0.

class Noise(GP):
    def __init__(self, nugget='nugget'):
        fa = dict(nugget=nugget)
        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['nugget'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)

        indices1 = tf.range(tf.shape(e['locs1'])[0])
        indices2 = tf.range(tf.shape(e['locs2'])[0]) + e['offset']
        C = tf.where(tf.equal(tf.expand_dims(indices1, -1), indices2), v['nugget'], 0.)
        return None, C

    def reg(self, p):
        return 0.

class Delta(GP):
    def __init__(self, dsill='dsill', axes=None):
        fa = dict(dsill=dsill)
        self.axes = axes
        super().__init__(fa, dict(pa_d2='per_axis_dist2'))

    def vars(self):
        return ppp(self.fa['dsill'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)

        if self.axes is not None:
            n = tf.shape(e['pa_d2'])[-1]
            mask = tf.math.bincount(self.axes, minlength=n, maxlength=n, dtype=tf.float32)
            d2 = tf.einsum('abc,c->ab', e['pa_d2'], mask)
        else:
            d2 = tf.reduce_sum(e['pa_d2'], axis=-1)

        return None, v['dsill'] * tf.cast(tf.equal(d2, 0.), tf.float32)

    def reg(self, p):
        return 0.

class Stack(GP):
    def __init__(self, parts: List[GP]):
        self.parts = parts
        super().__init__({}, dict(locs1='locs1', locs2='locs2', parts=parts))

    def vars(self):
        return [p for part in self.parts for p in part.vars()]

    def __add__(self, other):
        if isinstance(other, GP):
            return Stack(self.parts + [other])
    
    def call(self, p, e):
        MM, CC = zip(*e['parts'])
        MM = [M for M in MM if M is not None]
        CC = [C for C in CC if C is not None]
        M = tf.reduce_sum(MM, axis=0)
        C = tf.reduce_sum(CC, axis=0)
        return M, C

    def report(self, p):
        return ' '.join(part.report(p) for part in self.parts)

    def reg(self, p):
        return tf.reduce_sum([part.reg(p) for part in self.parts], axis=0)

class Product(GP):
    def __init__(self, parts: List[GP]):
        self.parts = parts
        super().__init__({}, dict(locs1='locs1', locs2='locs2', parts=parts))

    def vars(self):
        return [p for part in self.parts for p in part.vars()]

    def __mul__(self, other):
        if isinstance(other, GP):
            return Product(self.parts + [other])
    
    def call(self, p, e):
        MM, CC = zip(*e['parts'])
        MM = [M for M in MM if M is not None]
        CC = [C for C in CC if C is not None]
        M = tf.reduce_prod(MM, axis=0)
        C = tf.reduce_prod(CC, axis=0)
        return M, C

    def report(self, p):
        return ' '.join(part.report(p) for part in self.parts)

    def reg(self, p):
        return tf.reduce_sum([part.reg(p) for part in self.parts], axis=0)

# Gamma exponential covariance function.
@tf.custom_gradient
def safepow(x, a):
    y = tf.pow(x, a)
    def grad(dy):
        dx = tf.where(x <= 0.0, tf.zeros_like(x), dy * tf.pow(x, a-1))
        dx = unbroadcast(dx, x.shape)
        da = tf.where(x <= 0.0, tf.zeros_like(a), dy * y * tf.math.log(x))
        da = unbroadcast(da, a.shape)
        return dx, da
    return y, grad

def unbroadcast(x, shape):
    excess_rank = tf.maximum(0, len(x.shape) - len(shape))
    x = tf.reduce_sum(x, axis=tf.range(excess_rank))
    axes_that_are_one = tf.where(tf.equal(shape, 1))[:, 0]
    x = tf.reduce_sum(x, axis=axes_that_are_one, keepdims=True)
    return x

def gamma_exp(d2, gamma):
    return tf.exp(-safepow(tf.maximum(d2, 0.0), 0.5 * gamma))

class Observation(Op):

    def __init__(self,
        coefs: List,
        noise: GP
    ):
        self.coefs = coefs
        self.noise = noise
        super().__init__({}, self.noise)

    def vars(self):
        vv = [p for c in self.coefs for p in upp(c)]
        vv += self.noise.vars()
        return vv

    def __call__(self, p, e):
        """
        Dummy.
        """
        return 0.



import numpy as np
from geostat import gp, Model, Featurizer, NormalizingFeaturizer

def test_coefs1():
    # Create 100 random locations in a square centered on the origin.
    locs1 = np.random.uniform(-1., 1., [100, 2])

    # Initialize featurizer of location for trends.
    def trend_terms(x, y): return 1, y
    featurizer = Featurizer(trend_terms)
    covariance = gp.Trend(featurizer, beta='b') + gp.SquaredExponential(range=0.5, sill=0.01) + gp.Noise()

    # Generating GP.
    model1 = Model(
        latent = covariance,
        parameters = dict(b=[1.0, 0.5], nugget=0.1),
        verbose = True)

    # Generate data.
    vals1 = model1.generate(locs1).vals

    # Fit GP.
    model2 = Model(
        latent = covariance,
        parameters = dict(b=[1.0, 0.5], nugget=0.1),
        verbose = True).fit(locs1, vals1, iters=200)
    print()

    # Interpolate using GP.
    N = 2
    xx, yy = np.meshgrid(np.linspace(0, 1, N), np.linspace(0, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2])

    mean, var = model2.predict(locs2)

    target_mean = [1., 1., 1.5, 1.5]

    assert np.square(mean - target_mean).mean() < 0.1

def test_coefs2():
    # Create 500 random locations in a square centered on the origin.
    locs1 = np.random.uniform(-1., 1., [500, 2])

    # Initialize featurizer of location for trends.
    def trend_terms(x, y): return x, y, x*y
    featurizer = NormalizingFeaturizer(trend_terms, locs1)
    covariance = gp.Trend(featurizer, beta=[1., 'b1', 'b2', 'b3']) \
        + gp.SquaredExponential(sill=1.) + gp.Noise()

    # Generating GP.
    model1 = Model(
        latent = covariance,
        parameters = dict(b1=0.5, b2=0.5, b3=0.25, range=0.5, nugget=1.),
        verbose = True)

    # Generate data.
    vals1 = model1.generate(locs1).vals

    # Fit GP.
    model2 = Model(
        latent = covariance,
        parameters = dict(b1=0., b2=0., b3=0., range=1., nugget=0.5),
        verbose = True).fit(locs1, vals1, iters=200)
    print()

    # MCMC.
    model3 = Model(
        latent = covariance,
        parameters = dict(b1=0., b2=0., b3=0., range=1., nugget=0.5),
        verbose = True).mcmc(locs1, vals1, burnin=500, samples=500)

    # Interpolate using GP.
    N = 20
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2])

    mean, var = model2.predict(locs2)
    mean2, var2 = model2.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)

def test_explicit_trend():
    # Create 500 random locations in a square centered on the origin.
    locs1 = np.random.uniform(-1., 1., [500, 2])

    # Initialize featurizer of location for trends.
    def trend_terms(x, y): return x, y, x*y
    featurizer = NormalizingFeaturizer(trend_terms, locs1)
    covariance = gp.Trend(featurizer, beta='beta') \
        + gp.SquaredExponential(sill=1.) + gp.Noise()

    # Generating GP.
    model1 = Model(
        latent = covariance,
        parameters = dict(beta=[1., 0.5, 0.5, 0.25], range=0.5, nugget=1.),
        verbose = True)

    # Generate data.
    vals1 = model1.generate(locs1).vals

    # Fit GP.
    model2 = Model(
        latent = covariance,
        parameters = dict(beta=[0., 0., 0., 0.], range=1., nugget=0.5),
        verbose = True).fit(locs1, vals1, iters=200)
    print()

    # MCMC.
    model3 = Model(
        latent = covariance,
        parameters = dict(beta=[0., 0., 0., 0.], range=1., nugget=0.5),
        verbose = True).mcmc(locs1, vals1, burnin=500, samples=500)

    # Interpolate using GP.
    N = 20
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2])

    mean, var = model2.predict(locs2)
    mean2, var2 = model2.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)



import numpy as np
import tensorflow as tf
from geostat import gp, Model, Mesh, NormalizingFeaturizer


def test_int_sq_exp():

    np.random.seed(2)
    tf.random.set_seed(2)

    numsteps = 160
    xdim = 4.
    ydim = 0.5

    mesh = Mesh.from_bounds([0., 0., xdim, ydim], nx=numsteps)
    def trend_terms(x, y): return x, y, x*x, x*y, y*y
    featurizer = NormalizingFeaturizer(trend_terms, mesh.locations())

    mesh_vals = Model(
        gp.SquaredExponential(scale=[1., 1.]) + gp.Noise(nugget=1e-4),
        parameters = dict(range=0.1, sill=1.),
        verbose=True).generate(mesh.locations()).vals

    vmin, vmax = mesh_vals.min(), mesh_vals.max()
    meshx, meshy, mesh_vals_2d = mesh.slice(mesh_vals) # Each return value is a 2d array.

    int_vals = np.cumsum(mesh_vals_2d, axis=-1) / numsteps * xdim # Numerical integration on x-axis.

    sample_indices = np.random.choice(len(int_vals.ravel()), [500], replace=False)
    locs = np.stack([meshx.ravel()[sample_indices], meshy.ravel()[sample_indices]], axis=-1)
    vals = int_vals.ravel()[sample_indices]

    model = Model(
        gp.SquaredExponential(scale=[0., 1.], range='y_range') * \
        gp.IntSquaredExponential(axis=0, start=0., range='x_range') + \
        gp.Noise(nugget=1e-4),
        parameters = dict(x_range=1., y_range=1., sill=2.),
        verbose=True).fit(locs, vals, iters=2000, step_size=1e-1)

    assert np.allclose(
        [model.parameters[p] for p in ['x_range', 'y_range', 'sill']],
        [0.1, 0.1, 1.],
        rtol=0.5)

def test_int_exp():

    np.random.seed(2)
    tf.random.set_seed(2)

    numsteps = 160
    xdim = 4.
    ydim = 0.5

    mesh = Mesh.from_bounds([0., 0., xdim, ydim], nx=numsteps)
    def trend_terms(x, y): return x, y, x*x, x*y, y*y
    featurizer = NormalizingFeaturizer(trend_terms, mesh.locations())

    mesh_vals = Model(
        gp.SquaredExponential(scale=[0., 1.]) *
        gp.GammaExponential(scale=[1., 0.], sill=1., gamma=1.) + gp.Noise(nugget=1e-4),
        parameters = dict(range=0.1, sill=1.),
        verbose=True).generate(mesh.locations()).vals

    vmin, vmax = mesh_vals.min(), mesh_vals.max()
    meshx, meshy, mesh_vals_2d = mesh.slice(mesh_vals) # Each return value is a 2d array.

    int_vals = np.cumsum(mesh_vals_2d, axis=-1) / numsteps * xdim # Numerical integration on x-axis.

    sample_indices = np.random.choice(len(int_vals.ravel()), [500], replace=False)
    locs = np.stack([meshx.ravel()[sample_indices], meshy.ravel()[sample_indices]], axis=-1)
    vals = int_vals.ravel()[sample_indices]

    model = Model(
        gp.SquaredExponential(scale=[0., 1.], range='y_range') * \
        gp.IntExponential(axis=0, start=0., range='x_range') + \
        gp.Noise(nugget=1e-3),
        parameters = dict(x_range=1., y_range=1., sill=2.),
        verbose=True).fit(locs, vals, iters=2000, step_size=1e-1)

    assert np.allclose(
        [model.parameters[p] for p in ['x_range', 'y_range', 'sill']],
        [0.1, 0.1, 1.],
        rtol=0.5)

def test_wiener():

    np.random.seed(2)
    tf.random.set_seed(2)

    numsteps = 80
    xdim = 2.

    mesh = Mesh.from_bounds([0., 0., xdim, 1.], nx=numsteps)
    def trend_terms(x, y): return x, y, x*x, x*y, y*y
    featurizer = NormalizingFeaturizer(trend_terms, mesh.locations())

    mesh_vals = Model(
        gp.SquaredExponential(scale=[0., 1.], sill=1., range=0.3) *
        gp.SquaredExponential(scale=[1., 0.], sill=1., range=1e-4) + gp.Noise(nugget=1e-4),
        parameters = dict(),
        verbose=True).generate(mesh.locations()).vals

    vmin, vmax = mesh_vals.min(), mesh_vals.max()
    meshx, meshy, mesh_vals_2d = mesh.slice(mesh_vals) # Each return value is a 2d array.

    # Numerical integration on x-axis. Scaling is tricky.
    int_vals = np.cumsum(mesh_vals_2d, axis=-1) / np.sqrt(numsteps / xdim)

    sample_indices = np.random.choice(len(int_vals.ravel()), [500], replace=False)
    locs = np.stack([meshx.ravel()[sample_indices], meshy.ravel()[sample_indices]], axis=-1)
    vals = int_vals.ravel()[sample_indices]

    model = Model(
        gp.SquaredExponential(scale=[0., 1.]) * gp.Wiener(axis=0, start=0.) + gp.Noise(nugget=1e-4),
        parameters = dict(range=1., sill=2.),
        verbose=True).fit(locs, vals, iters=2000, step_size=1e-1)

    assert np.allclose(
        [model.parameters[p] for p in ['range', 'sill']],
        [0.3, 1.],
        rtol=0.5)



from dataclasses import dataclass
from typing import Dict

# Tensorflow is extraordinarily noisy. Catch warnings during import.
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import tensorflow as tf

@dataclass
class Op:
    """
    The `autoinputs` parameter contains a blob of upstream ops. The leaves
    in the blob are either the op itself or a string identifier. In the
    latter case, the string identifier should be present as a key in the
    `cache` that gets passed in.

    This parameter links ops together in a DAG. We walk the DAG for
    various reasons.

      - `run` calls the op and puts the output in `self.out`,
        after recursively doing this for autoinputs.
      - `gather_vars` recursively gathers variables from self and
        autoinputs.
    """
    fa: Dict[str, object] # Formal arguments.
    autoinputs: object # Blob of Ops and strings.

    def vars(self): # Parameters
        return []

    def gather_vars(self, cache=None):
        if cache is None: cache = {}
        if id(self) not in cache:
            vv = {v for op in tf.nest.flatten(self.autoinputs) if isinstance(op, Op) for v in op.gather_vars(cache)}
            cache[id(self)] = vv | set(self.vars())
        return cache[id(self)]

    def __call__(self, p, e):
        """
        `p` is a dict of model parameters.

        `e` is a blob of evaluated inputs from upstream ops and other
        things inserted by the caller.  Other values in `e` are supplied
        by the caller.
        """
        pass

    def run(self, cache, p):
        """
        If op has already been run, return result. Else:
            - Assemble inputs by recursively calling upstream ops.
            - Execute op by calling `__call__`.
            - Store result in cache.
        """

        def eval(op):
            """
            Evaluate `op`. If `op` is a string, look up its value.
            Otherwise execute it.
            """
            if isinstance(op, str):
                return cache[op]
            else:
                return op.run(cache, p)
       
        if id(self) not in cache:
            e = tf.nest.map_structure(lambda op: eval(op), self.autoinputs)
            cache[id(self)] = self(p, e)

            # Save the Op so that its ID remains unique.
            if '__save__' not in cache: cache['__save__'] = []
            cache['__save__'].append(self)

        return cache[id(self)]

    def __tf_tracing_type__(self, context):
        return SingletonTraceType(self)

class SingletonTraceType(tf.types.experimental.TraceType):
    """
    A trace type to override TF's default behavior, which is
    to treat dataclass-based onjects as dicts.
    """

    def __init__(self, thing):
       self.thing = thing
       pass

    def is_subtype_of(self, other):
       return type(other) is SingletonTraceType \
           and self.thing is other.thing

    def most_specific_common_supertype(self, others):
       return self if all(self == other for other in others) else None

    def __eq__(self, other):
       return isinstance(other, SingletonTraceType) and self.thing == other.thing

    def __hash__(self):
       return hash(id(self.thing))
